# NEON AI (TM) SOFTWARE, Software Development Kit & Application Framework
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2025 Neongecko.com Inc.
# Contributors: Daniel McKnight, Guy Daniels, Elon Gasper, Richard Leeds,
# Regina Bloomstine, Casimiro Ferreira, Andrii Pernatii, Kirill Hrymailo
# BSD-3 License
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import sys
import unittest
import shutil

sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))
from neon_utils.authentication_utils import *
from neon_utils.configuration_utils import NGIConfig

ROOT_DIR = os.path.dirname(os.path.realpath(__file__))
CRED_PATH = os.path.join(ROOT_DIR, "credentials")


class AuthUtilTests(unittest.TestCase):
    def setUp(self) -> None:
        config_path = os.path.join(ROOT_DIR, "configuration")
        self.old_local_conf = os.path.join(config_path, "old_local_conf.yml")
        self.ngi_local_conf = os.path.join(config_path, "ngi_local_conf.yml")
        shutil.copy(self.ngi_local_conf, self.old_local_conf)
        NGIConfig(os.path.splitext(os.path.basename(self.ngi_local_conf))[0],
                  os.path.dirname(self.ngi_local_conf),
                  force_reload=True)

    def tearDown(self) -> None:
        shutil.move(self.old_local_conf, self.ngi_local_conf)

    def test_get_git_token(self):
        try:
            token = find_neon_git_token("/tmp")
            self.assertIsInstance(token, str)
        except Exception as e:
            self.assertIsInstance(e, CredentialNotFoundError)

        token = find_neon_git_token(CRED_PATH)
        self.assertEqual(token, "github token goes here")

        os.environ["GITHUB_TOKEN"] = "test_gh_token"
        self.assertEqual(find_neon_git_token(), "test_gh_token")

    def test_get_aws_credentials(self):
        try:
            keys = find_neon_aws_keys("/tmp")
            self.assertEqual(list(keys.keys()), ["aws_access_key_id",
                                                 "aws_secret_access_key"])
        except Exception as e:
            self.assertIsInstance(e, CredentialNotFoundError)

        keys = find_neon_aws_keys(CRED_PATH)
        self.assertEqual(keys, {"aws_access_key_id": "FAKE_KEY_ID",
                                "aws_secret_access_key": "FAKE_SECRET/"})

        os.environ["AWS_ACCESS_KEY_ID"] = "test_aws_id"
        os.environ["AWS_SECRET_ACCESS_KEY"] = "test_aws_secret"
        self.assertEqual(find_neon_aws_keys(),
                         {"aws_access_key_id": "test_aws_id",
                          "aws_secret_access_key": "test_aws_secret"})

    def test_get_google_credentials(self):
        try:
            creds = find_neon_google_keys("/tmp")
            self.assertIsInstance(creds, dict)
        except Exception as e:
            self.assertIsInstance(e, CredentialNotFoundError)

        creds = find_neon_google_keys(CRED_PATH)
        self.assertEqual(list(creds.keys()),
                         ["type", "project_id", "private_key_id",
                          "private_key", "client_email", "client_id",
                          "auth_uri", "token_uri",
                          "auth_provider_x509_cert_url",
                          "client_x509_cert_url"])
        self.assertEqual(creds["private_key"],
                         "-----BEGIN PRIVATE KEY-----\nREDACTED\nREDACTED\n"
                         "REDACTED\n-----END PRIVATE KEY-----\n")

        os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = \
            os.path.join(CRED_PATH, "google.json")
        self.assertEqual(find_neon_google_keys(), creds)

    def test_get_wolfram_key(self):
        try:
            key = find_neon_wolfram_key("/tmp")
            self.assertIsInstance(key, str)
        except Exception as e:
            self.assertIsInstance(e, FileNotFoundError)

        key = find_neon_wolfram_key(CRED_PATH)
        self.assertEqual(key, "RED-ACTED")

        os.environ["WOLFRAM_APP_ID"] = "test_wa_id"
        self.assertEqual(find_neon_wolfram_key(), "test_wa_id")

    def test_get_alpha_vantage_key(self):
        try:
            key = find_neon_alpha_vantage_key("/tmp")
            self.assertIsInstance(key, str)
        except Exception as e:
            self.assertIsInstance(e, FileNotFoundError)

        key = find_neon_alpha_vantage_key(CRED_PATH)
        self.assertEqual(key, "Alpha-Vantage")

        os.environ["ALPHA_VANTAGE_KEY"] = "test_av_key"
        self.assertEqual(find_neon_alpha_vantage_key(), "test_av_key")

    def test_get_owm_key(self):
        try:
            key = find_neon_owm_key("/tmp")
            self.assertIsInstance(key, str)
        except Exception as e:
            self.assertIsInstance(e, FileNotFoundError)

        key = find_neon_owm_key(CRED_PATH)
        self.assertEqual(key, "OpenWeatherMap")

        os.environ["OWM_KEY"] = "test_owm_key"
        self.assertEqual(find_neon_owm_key(), "test_owm_key")

    def test_repo_is_neon_valid(self):
        self.assertTrue(repo_is_neon(
            "http://github.com/NeonGeckoCom/alerts.neon"))
        self.assertTrue(repo_is_neon(
            "https://github.com/NeonGeckoCom/caffeinewiz.neon"))
        self.assertTrue(repo_is_neon(
            "ssh://github.com/NeonGeckoCom/launcher.neon"))

        self.assertTrue(repo_is_neon(
            "https://github.com/neondaniel/speedtest.neon"))

        self.assertFalse(repo_is_neon(
            "https://github.com/mycroftai/skill-alarm"))
        self.assertFalse(repo_is_neon(
            "http://gitlab.com/neongecko/some-skill"))

    def test_repo_is_neon_invalid(self):
        with self.assertRaises(ValueError):
            repo_is_neon("https://github.com")
        with self.assertRaises(ValueError):
            repo_is_neon("not a url")
        with self.assertRaises(ValueError):
            repo_is_neon("")

    def test_build_new_auth_config(self):
        config = build_new_auth_config(CRED_PATH)
        self.assertEqual(set(config.keys()), {"github", "amazon", "wolfram",
                                              "google", "alpha_vantage",
                                              "owm"})
        for key in config.keys():
            self.assertIsInstance(config[key], dict)
            self.assertTrue(config[key])

        config = build_new_auth_config("/empty_dir")
        self.assertIsInstance(config, dict)
        for key in config.keys():
            if config[key] is not None:
                self.assertIsInstance(config[key], dict)
                for k, v in config[key].items():
                    self.assertIsInstance(k, str)
                    self.assertIsInstance(v, str)


if __name__ == '__main__':
    unittest.main()



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Framework
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2025 Neongecko.com Inc.
# Contributors: Daniel McKnight, Guy Daniels, Elon Gasper, Richard Leeds,
# Regina Bloomstine, Casimiro Ferreira, Andrii Pernatii, Kirill Hrymailo
# BSD-3 License
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import sys
import unittest
import socket
import threading

sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))
from neon_utils.logger import LOG
from neon_utils.socket_utils import *

TEST_DICT = {b"section 1": {"key1": "val1",
                            "key2": "val2"},
             "section 2": {"key_1": b"val1",
                           "key_2": f"val2"}}

TEST_DICT_B64 = b'IntiJ3NlY3Rpb24gMSc6IHsna2V5MSc6ICd2YWwxJywgJ2tleTInOiAndmFsMid9LCAnc2VjdGlvbiAyJzogeydrZXlfMSc6IGIndmFsMScsICdrZXlfMic6ICd2YWwyJ319Ig=='

TEST_SOCKET_ADDRESS = ('127.0.0.1', 8999)


class SocketUtilsTest(unittest.TestCase):

    def tcp_client(self):
        """Simple tcp socket client emitting just one message"""
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.connect(TEST_SOCKET_ADDRESS)
            s.sendall(TEST_DICT_B64)
            LOG.info('Client sent data')
            data = get_packet_data(socket=s, sequentially=True)
            self.assertEqual(data, TEST_DICT_B64)

    def test_01_dict_to_b64(self):
        b64_str = dict_to_b64(TEST_DICT)
        self.assertIsInstance(b64_str, bytes)
        self.assertTrue(len(b64_str) > 0)
        self.assertEqual(b64_str, TEST_DICT_B64)

    def test_02_b64_to_dict(self):
        result_dict = b64_to_dict(TEST_DICT_B64)
        self.assertIsInstance(result_dict, dict)
        self.assertTrue(len(list(result_dict)) > 0)
        self.assertEqual(result_dict, TEST_DICT)

    def test_03_get_packet_data(self):
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.bind(TEST_SOCKET_ADDRESS)
            s.listen()
            threading.Thread(target=self.tcp_client).start()
            conn, addr = s.accept()
            with conn:
                LOG.info(f'Connected by {addr}')
                data = get_packet_data(conn, sequentially=False)
                self.assertEqual(data, TEST_DICT_B64)
                conn.sendall(data)



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Framework
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2025 Neongecko.com Inc.
# Contributors: Daniel McKnight, Guy Daniels, Elon Gasper, Richard Leeds,
# Regina Bloomstine, Casimiro Ferreira, Andrii Pernatii, Kirill Hrymailo
# BSD-3 License
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import tracemalloc

from threading import Event, Thread
from typing import Optional
from ovos_utils.log import LOG
from ovos_utils.process_utils import ProcessStatus, ProcessState

_malloc_event = Event()
_malloc_thread = None


def start_systemd_service(service: callable, **kwargs):
    """
    Start a Neon Core module with systemd wrappers to report process lifecycle
    """
    try:
        import sdnotify
    except ImportError:
        LOG.exception(
            "sdnotify not installed! Starting service without systemd hooks"
        )
        service(**kwargs)
        return
    notifier = sdnotify.SystemdNotifier()

    def on_ready():
        notifier.notify("READY=1")
        notifier.notify("STATUS=Ready")

    def on_stopping():
        try:
            stop_malloc()
        except Exception as e:
            LOG.error(e)
        notifier.notify("STOPPING=1")
        notifier.notify("STATUS=Stopping")

    def on_error(err: str):
        if err.isnumeric():
            notifier.notify(f"ERRNO={err}")
        else:
            notifier.notify("ERRNO=1")

    def on_alive():
        notifier.notify("STATUS=Starting")

    def on_started():
        notifier.notify("STATUS=Started")

    service(
        ready_hook=on_ready,
        error_hook=on_error,
        stopping_hook=on_stopping,
        alive_hook=on_alive,
        started_hook=on_started,
        **kwargs,
    )


def start_health_check_server(
    service_status: ProcessStatus, port: int = 8000,
    status_callback: Optional[callable] = None
) -> Thread:
    """
    Starts an HTTP server to report the status of a module that implements a
    ProcessStatus object. A specified `status_callback` is expected to modify
    the `service_status` object as any returned value is ignored. If the
    `status_callback` raises an exception, the `service_status` will be
    updated here to set the error state.

    :param service_status: ProcessStatus object to get status from
    :param port: Port to run the health check server on
    :param status_callback: Optional callback to run on each health check
    """
    import json
    from http.server import BaseHTTPRequestHandler, HTTPServer

    class HealthCheckHandler(BaseHTTPRequestHandler):
        def log_message(self, format, *args):
            # Override logging to use LOG singleton
            LOG.debug(format % args)

        def do_GET(self):
            if self.server.status_callback is not None:
                try:
                    self.server.status_callback()
                except Exception as e:
                    self.server.service_status.set_error(str(e))
            if self.path == "/status":
                if self.server.service_status.state == ProcessState.NOT_STARTED:
                    resp_code = 503
                    content = "Service not started"
                elif self.server.service_status.state == ProcessState.STARTED:
                    resp_code = 200
                    content = "Starting"
                elif self.server.service_status.state == ProcessState.ALIVE:
                    resp_code = 200
                    content = "Starting"
                elif self.server.service_status.state == ProcessState.READY:
                    resp_code = 200
                    content = "Ready"
                elif self.server.service_status.state == ProcessState.STOPPING:
                    resp_code = 503
                    content = "Stopping"
                elif self.server.service_status.state == ProcessState.ERROR:
                    resp_code = 500
                    content = "Error"
                else:
                    resp_code = 500
                    content = f"Unknown state: {self.service_status.state}"

                self.send_response(resp_code)
                self.send_header("Content-Type", "application/json")
                self.end_headers()
                self.wfile.write(
                    json.dumps({"status": content}).encode("utf-8")
                )
            else:
                self.send_response(404)
                self.end_headers()

    server = HTTPServer(("0.0.0.0", port), HealthCheckHandler)
    server.service_status = service_status
    server.status_callback = status_callback
    thread = Thread(target=server.serve_forever, daemon=True)
    thread.start()
    LOG.info(f"Started health check endpoint at {server.server_address}")
    return thread


def start_malloc(
    config: dict = None, stack_depth: int = 1, force: bool = False
) -> bool:
    """
    Start malloc trace if configured
    :param config: dict Configuration
    :param stack_depth: depth to track memory usage
    :param force: if True, start tracemalloc regardless of configuration
    :returns: True if malloc started
    """
    if not config:
        from ovos_config.config import Configuration

        config = Configuration()
    if (
        force
        or config.get("debugging")
        and config["debugging"].get("tracemalloc")
    ):
        LOG.info("starting tracemalloc")
        tracemalloc.start(stack_depth)
        if config["debugging"].get("log_malloc"):
            interval_minutes = config["debugging"].get(
                "log_interval_minutes", 60
            )
            global _malloc_thread
            _malloc_thread = Thread(
                target=_log_malloc,
                args=((interval_minutes * 60),),
                daemon=True,
            )
            _malloc_thread.start()
        return True
    return False


def stop_malloc():
    """
    Stop tracemalloc logging if active
    """
    if _malloc_thread:
        LOG.debug("Stopping malloc logging")
        _malloc_event.set()
        _malloc_thread.join()


def snapshot_malloc() -> Optional[tracemalloc.Snapshot]:
    """
    Capture and return a tracemalloc Snapshot if tracemalloc is started
    """
    try:
        LOG.debug("Capturing malloc snapshot")
        return tracemalloc.take_snapshot()
    except RuntimeError:
        LOG.debug("No tracemalloc trace")
    except Exception as e:
        LOG.exception(e)
    return None


def print_malloc(
    snapshot: tracemalloc.Snapshot, limit: int = 8, filter_traces: bool = False
):
    """
    Log a malloc snapshot
    :param snapshot: Snapshot to evaluate
    :param limit: number of traces to log
    :param filter_traces: if True, remove importlib and unknown stack traces
    """
    LOG.debug("Processing snapshot")
    if filter_traces:
        snapshot = snapshot.filter_traces(
            (
                tracemalloc.Filter(False, "<frozen importlib._bootstrap>"),
                tracemalloc.Filter(
                    False, "<frozen importlib._bootstrap_external>"
                ),
                tracemalloc.Filter(False, "<unknown>"),
            )
        )
    top_stats = snapshot.statistics("traceback")

    LOG.info(f"Top {limit} memory users")
    for index, stat in enumerate(top_stats[:limit], 1):
        frame = stat.traceback[0]
        LOG.info(
            f"#{index}: {frame.filename}:{frame.lineno}: "
            f"{stat.size / 1048576} MiB"
        )
        for frame in stat.traceback[1:]:
            LOG.debug(f"{frame.filename}:{frame.lineno}")

    other = top_stats[limit:]
    if other:
        size = sum(stat.size for stat in other)
        LOG.info(f"{len(other)} other:{size / 1048576} MiB")
    total = sum(stat.size for stat in top_stats)
    LOG.info(f"Total allocated size: {total / 1048576} MiB")


def _log_malloc(interval_seconds: int):
    _malloc_event.clear()
    while not _malloc_event.wait(interval_seconds):
        print_malloc(snapshot_malloc(), filter_traces=True)
    LOG.info("Stopping malloc logging")



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Framework
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2025 Neongecko.com Inc.
# Contributors: Daniel McKnight, Guy Daniels, Elon Gasper, Richard Leeds,
# Regina Bloomstine, Casimiro Ferreira, Andrii Pernatii, Kirill Hrymailo
# BSD-3 License
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import sys
import unittest
from unittest.mock import patch

sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))
from neon_utils.packaging_utils import *


class PackagingUtilTests(unittest.TestCase):
    def test_get_neon_core_version(self):
        version = get_neon_core_version()
        self.assertIsInstance(version, str)
        self.assertGreaterEqual(len(version.split('.')), 2)

    def test_get_core_root(self):
        try:
            core_dir = get_core_root()
            self.assertIsInstance(core_dir, str)
            self.assertTrue(os.path.isdir(os.path.join(core_dir, "mycroft")))
        except FileNotFoundError:
            with self.assertRaises(FileNotFoundError):
                get_core_root()

    def test_get_neon_core_root(self):
        try:
            core_dir = get_neon_core_root()
            self.assertIsInstance(core_dir, str)
            self.assertTrue(os.path.isdir(os.path.join(core_dir, "util")))
        except FileNotFoundError:
            with self.assertRaises(FileNotFoundError):
                get_neon_core_root()

    def test_get_mycroft_core_root(self):
        try:
            core_dir = get_mycroft_core_root()
            self.assertIsInstance(core_dir, str)
            self.assertTrue(os.path.isdir(os.path.join(core_dir, "mycroft")))
        except FileNotFoundError:
            with self.assertRaises(FileNotFoundError):
                get_mycroft_core_root()

    def test_get_package_version_spec(self):
        ver = get_package_version_spec("ovos_utils")
        self.assertIsInstance(ver, str)

        with self.assertRaises(ModuleNotFoundError):
            get_package_version_spec("neon-stt-fake-test-package")

        with self.assertRaises(ModuleNotFoundError):
            get_package_version_spec("mycroft")

    def test_parse_version_string(self):
        major, minor, patch, alpha = parse_version_string("21.5")
        self.assertEqual(major, 21)
        self.assertEqual(minor, 5)
        self.assertEqual(patch, 0)
        self.assertEqual(alpha, None)

        major, minor, patch, alpha = parse_version_string("21.5.1")
        self.assertEqual(major, 21)
        self.assertEqual(minor, 5)
        self.assertEqual(patch, 1)
        self.assertEqual(alpha, None)

        major, minor, patch, alpha = parse_version_string("21.5.1a0")
        self.assertEqual(major, 21)
        self.assertEqual(minor, 5)
        self.assertEqual(patch, 1)
        self.assertEqual(alpha, 0)

        major, minor, patch, alpha = parse_version_string("21.5.2post2")
        self.assertEqual(major, 21)
        self.assertEqual(minor, 5)
        self.assertEqual(patch, 2)
        self.assertEqual(alpha, 2)

    def test_get_packaged_core_version(self):
        try:
            ver = get_packaged_core_version()
            self.assertIsInstance(ver, str)
            self.assertGreaterEqual(len(ver.split('.')), 2)
        except ImportError:
            with self.assertRaises(ImportError):
                get_packaged_core_version()

    def test_get_package_dependencies(self):
        self_deps = get_package_dependencies("neon-utils")
        requirements_file = join(os.path.dirname(os.path.dirname(__file__)),
                                 "requirements", "requirements.txt")
        with open(requirements_file) as f:
            spec_requirements = f.read().split('\n')
        spec_requirements = [r for r in spec_requirements
                             if r and not r.startswith('#')]
        # Version specs aren't order-dependent, so they can't be compared
        self.assertEqual(len(self_deps), len(spec_requirements))
        with self.assertRaises(ModuleNotFoundError):
            get_package_dependencies("fakeneongeckopackage")

    def test_build_skill_spec(self):
        import json
        from neon_utils.packaging_utils import build_skill_spec
        test_dir = os.path.join(os.path.dirname(__file__), "test_skill_json")

        with self.assertRaises(FileNotFoundError):
            build_skill_spec(os.path.join(test_dir, "notADirectory"))
        with self.assertRaises(FileNotFoundError):
            build_skill_spec(__file__)

        skill_spec = build_skill_spec(test_dir)
        self.assertIsInstance(skill_spec, dict)
        # Patch params otherwise read from .git directory
        skill_spec["authorname"] = "NeonGeckoCom"
        skill_spec["skillname"] = "skill-alerts"
        skill_spec["url"] = "https://github.com/NeonGeckoCom/skill-alerts"
        with open(join(test_dir, "skill.json")) as f:
            valid_spec = json.load(f)
        self.assertEqual(valid_spec, skill_spec, f"actual={skill_spec}")

    # TODO: Actually validate exception cases? DM
        
    def test_install_packages_from_pip(self):
        import pip
        from neon_utils.packaging_utils import install_packages_from_pip

        with patch.object(pip, 'main', return_value=0) as mock_method:
            test_result = install_packages_from_pip("neon-utils", ["pip-install-test"])
        
            args, kwargs = mock_method.call_args
            self.assertEqual(0, test_result)
            mock_method.assert_called_once_with(['install', '-r', args[0][2], '-c', args[0][4]])

            with open(args[0][2], "r", encoding="utf8") as f:
                line = f.readline()
                self.assertEqual("pip-install-test\n", line)

            mock_method.reset_mock()
            test_result = install_packages_from_pip("neon-utils", ["pip-install-test", "pip-install-another"])

            self.assertEqual(0, test_result)
            mock_method.assert_called_once()



if __name__ == '__main__':
    unittest.main()



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Framework
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2025 Neongecko.com Inc.
# Contributors: Daniel McKnight, Guy Daniels, Elon Gasper, Richard Leeds,
# Regina Bloomstine, Casimiro Ferreira, Andrii Pernatii, Kirill Hrymailo
# BSD-3 License
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import sys
import unittest
from unittest import skip

import pika

from threading import Thread
from time import time

sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))
from neon_utils.socket_utils import dict_to_b64
from neon_utils.mq_utils import *

ROOT_DIR = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
TEST_PATH = os.path.join(ROOT_DIR, "tests", "ccl_files")

INPUT_CHANNEL = str(time())
OUTPUT_CHANNEL = str(time())


class TestMQConnector(MQConnector):
    def __init__(self, config: dict, service_name: str, vhost: str):
        super().__init__(config, service_name)
        self.vhost = vhost

    @staticmethod
    def respond(channel, method, _, body):
        request = b64_to_dict(body)
        response = dict_to_b64({"message_id": request["message_id"],
                                "success": True,
                                "request_data": request["data"]})
        reply_channel = request.get("routing_key") or OUTPUT_CHANNEL
        channel.queue_declare(queue=reply_channel)
        channel.basic_publish(exchange='',
                              routing_key=reply_channel,
                              body=response,
                              properties=pika.BasicProperties(expiration='1000'))
        channel.basic_ack(delivery_tag=method.delivery_tag)


@skip("MQ Utils have moved to `neon-mq-connector`")
class MqUtilTests(unittest.TestCase):
    test_connector = None

    @classmethod
    def setUpClass(cls) -> None:
        from neon_utils.mq_utils import _default_mq_config
        vhost = "/neon_testing"
        cls.test_connector = TestMQConnector(config=_default_mq_config,
                                             service_name="mq_handler",
                                             vhost=vhost)
        cls.test_connector.register_consumer("neon_utils_test", vhost,
                                             INPUT_CHANNEL,
                                             cls.test_connector.respond,
                                             auto_ack=False)
        cls.test_connector.run_consumers()

    @classmethod
    def tearDownClass(cls) -> None:
        cls.test_connector.stop_consumers()

    def test_send_mq_request_valid(self):
        request = {"data": time()}
        response = send_mq_request("/neon_testing", request, INPUT_CHANNEL)
        self.assertIsInstance(response, dict)
        self.assertTrue(response["success"])
        self.assertEqual(response["request_data"], request["data"])

    def test_send_mq_request_spec_output_channel_valid(self):
        request = {"data": time()}
        response = send_mq_request("/neon_testing", request, INPUT_CHANNEL, OUTPUT_CHANNEL)
        self.assertIsInstance(response, dict)
        self.assertTrue(response["success"])
        self.assertEqual(response["request_data"], request["data"])

    def test_multiple_mq_requests(self):
        responses = dict()
        processes = []

        def check_response(name: str):
            request = {"data": time()}
            response = send_mq_request("/neon_testing", request, INPUT_CHANNEL)
            self.assertIsInstance(response, dict)
            if not isinstance(response, dict):
                responses[name] = {'success': False,
                                   'reason': 'Response is not a dict',
                                   'response': response}
                return
            if not response.get("success"):
                responses[name] = {'success': False,
                                   'reason': 'Response success flag not true',
                                   'response': response}
                return
            if response.get("request_data") != request["data"]:
                responses[name] = {'success': False,
                                   'reason': "Response data doesn't match request",
                                   'response': response}
                return
            responses[name] = {'success': True}

        for i in range(8):
            p = Thread(target=check_response, args=(str(i),))
            p.start()
            processes.append(p)

        for p in processes:
            p.join(60)

        self.assertEqual(len(processes), len(responses))
        for resp in responses.values():
            self.assertTrue(resp['success'], resp.get('reason'))

    def test_send_mq_request_invalid_vhost(self):
        with self.assertRaises(ValueError):
            send_mq_request("invalid_endpoint", {}, "test", "test", timeout=5)

    # def test_get_mq_response_error_in_handler(self):
    #     # TODO: Troubleshoot this DM
    #     request = {"fail": True}
    #     response = get_mq_response("/neon_testing", request, INPUT_CHANNEL, OUTPUT_CHANNEL)
    #     self.assertIsInstance(response, dict)
    #     self.assertFalse(response)


if __name__ == '__main__':
    unittest.main()



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Framework
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2025 Neongecko.com Inc.
# Contributors: Daniel McKnight, Guy Daniels, Elon Gasper, Richard Leeds,
# Regina Bloomstine, Casimiro Ferreira, Andrii Pernatii, Kirill Hrymailo
# BSD-3 License
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import datetime

from abc import abstractmethod
from threading import Thread
from time import time
from ovos_bus_client import Message
from ovos_utils.log import LOG

from neon_utils.message_utils import get_message_user, dig_for_message
from neon_utils.skills.neon_skill import NeonSkill


class KioskSkill(NeonSkill):
    def __init__(self, *args, **kwargs):
        NeonSkill.__init__(self, *args, **kwargs)
        self._active_users = dict()

    @property
    def timeout_seconds(self) -> int:
        """
        Time in seconds to wait for a user response before timing out
        """
        return 60

    @property
    @abstractmethod
    def greeting_dialog(self) -> str:
        """
        Specify a dialog to speak on interaction start
        """

    @property
    @abstractmethod
    def goodbye_dialog(self) -> str:
        """
        Specify a dialog to speak when an interaction ends cleanly
        """

    @property
    @abstractmethod
    def timeout_dialog(self) -> str:
        """
        Specify a dialog to speak when an interaction ends after some timeout
        """

    @property
    @abstractmethod
    def error_dialog(self) -> str:
        """
        Specify a dialog to speak on unhandled errors
        """

    def start_interaction(self, message: Message):
        """
        Start an interaction session. This may be triggered by intent,
        proximity, button press, etc.
        :param message: Message associated with request
        """
        if self.setup_new_interaction(message):
            user = get_message_user(message)
            self.speak_dialog(self.greeting_dialog)
            self._active_users[user] = time()
            self.handle_new_interaction(message)
            self._active_users[user] = time()
            self._timeout_timer(message)

    @abstractmethod
    def setup_new_interaction(self, message: Message) -> bool:
        """
        Override to include skill-specific actions on first user interaction.
        This is the first action that could prompt user to input language, etc.
        :param message: Message associated with start request
        :returns: True if user interaction is supported
        """
        return True

    @abstractmethod
    def handle_new_interaction(self, message: Message):
        """
        Override to interact with the user after the greeting message has been
        spoken.
        :param message: Message associated with start request
        """
        pass

    def end_interaction(self, message: Message):
        """
        Handle ending an interaction session. This is called if a user requests
        'stop' or if the skill decides to end interaction.
        :param message: Message associated with request
        """
        self.handle_end_interaction(message)
        self.speak_dialog(self.goodbye_dialog)
        user = get_message_user(message)
        self._active_users.pop(user)
        self.cancel_scheduled_event(f'{self.skill_id}:timeout_{user}')

    @abstractmethod
    def handle_end_interaction(self, message: Message):
        """
        Override to do any skill-specific cleanup when a user interaction is
        completed.
        :param message: Message associated with request triggering end
        """
        pass

    @abstractmethod
    def handle_user_utterance(self, message: Message):
        """
        Handle any input from a user interacting with the kiosk.
        :param message: Message associated with user utterance
        """
        self.handle_error(message)

    def handle_error(self, message: Message):
        """
        Error handler to speak a custom dialog file when an error occurs during
        user interaction.
        """
        self.speak_dialog(self.error_dialog, message=message)

    def converse(self, message: Message = None):
        user = get_message_user(message)
        if user in self._active_users:
            if self.voc_match(message.data.get('utterance'), 'stop'):
                self.bus.emit(message.forward("mycroft.stop"))
                return True
            LOG.debug(f"Consuming utterance from {user}")
            Thread(target=self._handle_converse, args=(message,),
                   daemon=True).start()
            return True
        return False

    def _handle_converse(self, message: Message):
        """
        Handle actual converse interaction in a thread so the skills service can
        get a `handled` response immediately
        """
        user = get_message_user(message)
        self.handle_user_utterance(message)
        self._active_users[user] = time()
        self._timeout_timer(message)

    def _timeout_timer(self, message: Message):
        """
        Schedule some event to handle no response from a particular user
        """
        user = get_message_user(message)
        timeout_time = datetime.datetime.fromtimestamp(
            self._active_users[user]) + \
            datetime.timedelta(seconds=self.timeout_seconds)
        event_name = f'{self.skill_id}:timeout_{user}'
        self.schedule_event(self._handle_timeout, timeout_time, message.data,
                            event_name, message.context)

    def _handle_timeout(self, message):
        """
        Handler for when a user hasn't responded within the specified timeout
        """
        user = get_message_user(message)
        if user in self._active_users:
            LOG.info(f"Response timeout for: {user}")
            self._active_users.pop(user)
            if self.timeout_dialog:
                self.speak_dialog(self.timeout_dialog)

    def stop(self):
        message = dig_for_message()
        user = get_message_user(message) if message else "local"
        if user in self._active_users:
            self.end_interaction(message)

    def _on_event_error(self, error, message, *args, **kwargs):
        """
        Override error handling to speak custom exception for active sessions
        """
        user = get_message_user(message)
        if user in self._active_users:
            LOG.exception(error)
            self.handle_error(message)
        else:
            super()._on_event_error(error, message, *args, **kwargs)



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Framework
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2025 Neongecko.com Inc.
# Contributors: Daniel McKnight, Guy Daniels, Elon Gasper, Richard Leeds,
# Regina Bloomstine, Casimiro Ferreira, Andrii Pernatii, Kirill Hrymailo
# BSD-3 License
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import unittest
import tempfile
import shutil
import subprocess

from os.path import join, dirname
from mock import patch


class SkillUtilTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        """Set up test fixtures, downloading the test skill repository"""
        cls.test_dir = tempfile.mkdtemp(prefix="skill_util_tests_")

        # Set up setuptools test repository
        cls.setuptools_repo_url = (
            "https://github.com/NeonGeckoCom/skill-caffeinewiz.git"
        )
        cls.setuptools_skill_branch = "0.3.1"
        cls.setuptools_skill_repo_path = os.path.join(
            cls.test_dir, "skill-caffeinewiz"
        )
        subprocess.run(
            [
                "git",
                "clone",
                "--branch",
                cls.setuptools_skill_branch,
                "--depth",
                "1",
                cls.setuptools_repo_url,
                cls.setuptools_skill_repo_path,
            ],
            check=True,
            capture_output=True,
            text=True,
            timeout=30,
        )
        cls.setup_py_path = os.path.join(
            cls.setuptools_skill_repo_path, "setup.py"
        )

        # Set up poetry test repository
        cls.poetry_repo_url = (
            "https://github.com/OscillateLabsLLC/skill-homeassistant.git"
        )
        cls.poetry_skill_branch = "v0.5.1"
        cls.poetry_skill_repo_path = os.path.join(
            cls.test_dir, "skill-homeassistant"
        )
        subprocess.run(
            [
                "git",
                "clone",
                "--branch",
                cls.poetry_skill_branch,
                "--depth",
                "1",
                cls.poetry_repo_url,
                cls.poetry_skill_repo_path,
            ],
            check=True,
            capture_output=True,
            text=True,
            timeout=30,
        )
        cls.pyproject_toml_path = os.path.join(
            cls.poetry_skill_repo_path, "pyproject.toml"
        )

    @classmethod
    def tearDownClass(cls):
        """Clean up test fixtures"""
        if hasattr(cls, "test_dir") and os.path.exists(cls.test_dir):
            shutil.rmtree(cls.test_dir)

    @patch("neon_utils.skill_utils._get_skill_data_poetry")
    @patch("neon_utils.skill_utils._get_skill_data_setuptools")
    def test_get_skill_metadata(self, mock_setuptools, mock_poetry):
        """Test that get_skill_metadata calls appropriate internal methods based on file existence"""
        from neon_utils.skill_utils import get_skill_metadata

        # Set up mock return values
        mock_poetry_data = {"name": "skill-homeassistant", "source": "poetry"}
        mock_setuptools_data = {
            "name": "skill-caffeinewiz",
            "source": "setuptools",
        }
        mock_poetry.return_value = mock_poetry_data
        mock_setuptools.return_value = mock_setuptools_data

        # Test poetry skill (homeassistant)
        result_poetry = get_skill_metadata(self.poetry_skill_repo_path)
        mock_poetry.assert_called_with(self.pyproject_toml_path)
        mock_setuptools.assert_not_called()
        self.assertEqual(result_poetry['name'], mock_poetry_data['name'])

        # Reset mocks
        mock_poetry.reset_mock()
        mock_setuptools.reset_mock()

        # Test setuptools skill (caffeinewiz)
        result_setuptools = get_skill_metadata(self.setuptools_skill_repo_path)
        mock_setuptools.assert_called_with(self.setup_py_path)
        mock_poetry.assert_not_called()
        self.assertEqual(result_setuptools['name'], mock_setuptools_data['name'])
        # Test FileNotFoundError for non-existent directory
        with self.assertRaises(FileNotFoundError):
            get_skill_metadata("/non/existent/directory")

    def test_get_skill_data_poetry(self):
        from neon_utils.skill_utils import _get_skill_data_poetry

        # Test with real skill-homeassistant pyproject.toml file
        self.assertTrue(
            os.path.exists(self.pyproject_toml_path),
            f"pyproject.toml not found at {self.pyproject_toml_path}",
        )

        skill_metadata = _get_skill_data_poetry(self.pyproject_toml_path)
        self.assertIsInstance(skill_metadata, dict)
        self.assertEqual(skill_metadata.get("name"), "skill-homeassistant")
        self.assertEqual(
            skill_metadata.get("package_name"), "skill-homeassistant"
        )
        self.assertEqual(skill_metadata.get("pip_spec"), "skill-homeassistant")
        self.assertEqual(skill_metadata.get("version"), "0.5.1")
        self.assertEqual(skill_metadata.get("license"), "Apache-2.0")
        self.assertIn("Mike Gray", skill_metadata.get("author")[0])
        self.assertEqual(skill_metadata.get("title"), "skill-homeassistant")
        self.assertIsInstance(skill_metadata.get("tags"), list)
        self.assertIn("neon", skill_metadata.get("tags"))
        self.assertIsInstance(
            skill_metadata.get("requirements", {}).get("python"), list
        )
        self.assertGreaterEqual(
            len(skill_metadata["requirements"]["python"]), 1
        )

        # Test params from README
        self.assertIsInstance(skill_metadata["summary"], str)

        # Test FileNotFoundError for non-existent file
        with self.assertRaises(FileNotFoundError):
            _get_skill_data_poetry("non_existent_file.toml")

    def test_get_skill_data_setuptools(self):
        from neon_utils.skill_utils import _get_skill_data_setuptools

        # Test with real skill-caffeinewiz setup.py file
        self.assertTrue(
            os.path.exists(self.setup_py_path),
            f"setup.py not found at {self.setup_py_path}",
        )

        skill_metadata = _get_skill_data_setuptools(self.setup_py_path)
        self.assertIsInstance(skill_metadata, dict)
        self.assertEqual(skill_metadata.get("authorname"), "NeonGeckoCom")
        self.assertEqual(skill_metadata.get("skillname"), "skill-caffeinewiz")
        self.assertEqual(
            skill_metadata["package_name"], "neon-skill-caffeinewiz"
        )
        self.assertEqual(skill_metadata["pip_spec"], "neon-skill-caffeinewiz")
        self.assertIn("BSD-3", skill_metadata["license"])
        self.assertIsInstance(skill_metadata["author"], str)
        self.assertEqual(skill_metadata["version"], "0.3.1")
        self.assertIsInstance(skill_metadata["url"], str)
        self.assertEqual(skill_metadata["name"], "neon-skill-caffeinewiz")
        self.assertEqual(skill_metadata["title"], "neon-skill-caffeinewiz")
        self.assertIn("Neongecko", skill_metadata["credits"])
        self.assertIsInstance(skill_metadata["requirements"]["python"], list)
        self.assertGreaterEqual(
            len(skill_metadata["requirements"]["python"]), 1
        )

        # Test params from README
        self.assertIsInstance(skill_metadata["summary"], str)

        # Test FileNotFoundError for non-existent file
        with self.assertRaises(FileNotFoundError):
            _get_skill_data_setuptools("non_existent_setup.py")

    def test_get_skill_data_readme(self):
        from neon_utils.skill_utils import _get_skill_data_readme

        valid_readme_file = join(dirname(__file__), "test_skill_json", "README.md")

        with open(valid_readme_file, 'r') as f:
            readme_contents = f.read()

        skill_metadata = _get_skill_data_readme(readme_contents)
        self.assertIsInstance(skill_metadata, dict)
        self.assertIsInstance(skill_metadata['examples'], list)
        self.assertIsInstance(skill_metadata['incompatible_skills'], list)
        self.assertIsInstance(skill_metadata['categories'], list)
        self.assertIsInstance(skill_metadata['tags'], list)
        self.assertIsInstance(skill_metadata['credits'], list)

        self.assertIsInstance(skill_metadata['summary'], str)
        self.assertIsInstance(skill_metadata['description'], str)

        self.assertIsInstance(skill_metadata['title'], str)
        self.assertIsInstance(skill_metadata['icon'], str)

        # Test invalid README contents
        with self.assertRaises(ValueError):
            _get_skill_data_readme("")

if __name__ == "__main__":
    unittest.main()



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Framework
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2025 Neongecko.com Inc.
# Contributors: Daniel McKnight, Guy Daniels, Elon Gasper, Richard Leeds,
# Regina Bloomstine, Casimiro Ferreira, Andrii Pernatii, Kirill Hrymailo
# BSD-3 License
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from neon_utils.logger import LOG


def search_convo_dict(id_to_keywords: dict, search_term: str,
                      max_matches_to_return: int = 5, handle_typos: bool = True) -> list:
    """
    Searches the passed id_to_keywords dict for id's matching the search_term
    :param id_to_keywords: dict of id to list of keywords to match
    :param search_term: string search term to locate
    :param max_matches_to_return: max number of sorted results to return
    :param handle_typos: boolean to allow inexact string matches
    :return: list of matched IDs
    """
    def _return_close_values(source_dict, to_find, num_of_return, typo_check=True) -> list:
        import difflib
        allow_diffs = 1  # if len(to_find) < 4 else 2

        perfect_match = [i for i, v in source_dict.items() if (isinstance(v, str) and to_find == v)
                         or (isinstance(v, list) and [x for x in v if to_find == x])]
        contains_in = [k for k, i in source_dict.items() if to_find in i or (isinstance(i, str) and i in to_find)
                       or (isinstance(i, list) and [x for x in i if x in to_find or to_find in x])]
        close_match = list(set(k for c in difflib.get_close_matches(to_find, source_dict.values())
                               for k, v in source_dict.items() if v == c))

        def check_diffs(i, k):
            len_i = len(i)
            len_tf = len(to_find)
            if abs(len_i - len_tf) > allow_diffs:
                return
            counter_i, counter_tf = 0, 0
            diff_count = 0
            while counter_i < len_i and counter_tf < len_tf:
                if to_find[counter_tf] != i[counter_i]:
                    diff_count += 1
                    if diff_count > allow_diffs:
                        break
                    if len_tf > len_i:
                        counter_tf += 1
                    elif len_i > len_tf:
                        counter_i += 1
                    else:
                        counter_i += 1
                        counter_tf += 1
                else:
                    counter_i += 1
                    counter_tf += 1

            if diff_count <= allow_diffs:
                typos_check.append(k)

        typos_check = []
        if typo_check:
            for k, i in source_dict.items():
                if isinstance(i, str):
                    check_diffs(i, k)
                elif isinstance(i, list):
                    for x in i:
                        if k not in typos_check:
                            check_diffs(x, k)

        LOG.debug(f"perfect match - {perfect_match}")
        LOG.debug(f'contains part - {contains_in}')
        LOG.debug(f"close_match - {close_match}")
        if typo_check:
            LOG.debug(f"typo check {allow_diffs} characters diff - {typos_check}")

        if not close_match and not contains_in and not perfect_match and not typos_check:
            return []
        try:
            res_tmp = typos_check
            res_tmp.extend([i for i in close_match if i not in typos_check])
            res_tmp.extend([i for i in contains_in if i not in close_match])
            LOG.debug(f'res_tmp is {res_tmp}')
        except Exception as x:
            LOG.debug(x)
            res_tmp = []

        if perfect_match:
            if len(perfect_match) == num_of_return:
                return perfect_match
            else:
                need_add = num_of_return - len(perfect_match)
                LOG.debug(len(perfect_match))
                LOG.debug(need_add)
                if need_add < 0:
                    return perfect_match[:num_of_return]

                res = perfect_match
                if res_tmp:
                    res.extend([i for i in res_tmp if i not in perfect_match])
                    LOG.debug(res)
        else:
            res = res_tmp

        if not res:
            raise ValueError("No results matched!")
        return res[:num_of_return]

    fin_res = []
    from collections import Counter
    import operator
    if not search_term:
        return []
    if " " in search_term:
        for i in search_term.split(" "):
            fin_res.extend(_return_close_values(id_to_keywords, i, max_matches_to_return, handle_typos))
        LOG.debug(fin_res)
        tmp_f = Counter(fin_res)
        LOG.debug(tmp_f)
        tmp = sorted(tmp_f.items(), key=operator.itemgetter(1), reverse=True)
        LOG.debug(tmp)
        return [i for i, m in tmp][:max_matches_to_return]
    return _return_close_values(id_to_keywords, search_term, max_matches_to_return, handle_typos)



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Framework
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2025 Neongecko.com Inc.
# Contributors: Daniel McKnight, Guy Daniels, Elon Gasper, Richard Leeds,
# Regina Bloomstine, Casimiro Ferreira, Andrii Pernatii, Kirill Hrymailo
# BSD-3 License
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import json
import unittest

from os import remove
from os.path import join, dirname, isfile
from shutil import copy
from time import time, sleep
from unittest.mock import patch

valid_config = {}
valid_headers = {}


class HanaUtilTests(unittest.TestCase):
    test_server = "https://hana.neonaibeta.com"
    test_path = join(dirname(__file__), "hana_test.json")

    def tearDown(self) -> None:
        global valid_config
        global valid_headers
        import neon_utils.hana_utils
        if isfile(self.test_path):
            remove(self.test_path)
        if neon_utils.hana_utils._client_config:
            valid_config = neon_utils.hana_utils._client_config
        if neon_utils.hana_utils._headers:
            valid_headers = neon_utils.hana_utils._headers
        neon_utils.hana_utils._client_config = {}
        neon_utils.hana_utils._headers = {}

    @patch("neon_utils.hana_utils._get_client_config_path")
    def test_request_backend(self, config_path):
        config_path.return_value = self.test_path

        # Use a valid config and skip extra auth
        import neon_utils.hana_utils
        neon_utils.hana_utils._client_config = valid_config
        neon_utils.hana_utils._headers = valid_headers
        from neon_utils.hana_utils import request_backend
        resp = request_backend("/neon/get_response",
                               {"lang_code": "en-us",
                                "utterance": "who are you",
                                "user_profile": {}}, self.test_server)
        self.assertEqual(resp['lang_code'], "en-us")
        self.assertIsInstance(resp['answer'], str)

        # Test expired/invalid token
        old_token_path = join(dirname(__file__), "outdated_hana_token.json")
        copy(old_token_path, self.test_path)
        with open(self.test_path, 'r') as f:
            old_contents = f.read()
        neon_utils.hana_utils._client_config = {}
        neon_utils.hana_utils._headers = {}

        # Request generates an updated token
        resp = request_backend("/neon/get_response",
                               {"lang_code": "en-us",
                                "utterance": "who are you",
                                "user_profile": {}}, self.test_server)
        self.assertEqual(resp['lang_code'], "en-us")
        self.assertIsInstance(resp['answer'], str)

        # New token is created at expected path
        self.assertTrue(isfile(self.test_path))
        with open(self.test_path, 'r') as f:
            new_contents = f.read()
        self.assertNotEqual(new_contents, old_contents)

        # TODO: Test token refresh fails, old token is removed
        # TODO: Test invalid route, invalid request data

    @patch("neon_utils.hana_utils._get_client_config_path")
    @patch("neon_utils.hana_utils._refresh_token")
    def test_request_backend_refresh_token(self, refresh_token, config_path):
        config_path.return_value = self.test_path

        import neon_utils.hana_utils
        from neon_utils.hana_utils import request_backend
        neon_utils.hana_utils.set_default_backend_url(self.test_server)
        neon_utils.hana_utils._init_client(self.test_server)
        real_client_config = neon_utils.hana_utils._client_config
        neon_utils.hana_utils._client_config['expiration'] = time() + 29
        neon_utils.hana_utils._refresh_token = refresh_token
        resp = request_backend("/neon/get_response",
                               {"lang_code": "en-us",
                                "utterance": "how are you",
                                "user_profile": {}}, self.test_server)
        self.assertEqual(resp['lang_code'], "en-us")
        self.assertIsInstance(resp['answer'], str)
        refresh_token.assert_called_once_with(self.test_server,
                                              ssl_verify=True)

        neon_utils.hana_utils._client_config = real_client_config

    @patch("neon_utils.hana_utils._get_client_config_path")
    def test_00_get_token(self, config_path):
        config_path.return_value = self.test_path
        from neon_utils.hana_utils import _get_token

        # Test valid request
        _get_token(self.test_server)
        from neon_utils.hana_utils import _client_config
        self.assertTrue(isfile(self.test_path))
        with open(self.test_path) as f:
            credentials_on_disk = json.load(f)
        self.assertEqual(credentials_on_disk, _client_config)
        # TODO: Test invalid request, rate-limited request

    @patch("neon_utils.hana_utils._get_client_config_path")
    @patch("neon_utils.hana_utils._get_token")
    def test_refresh_token(self, get_token, config_path):
        config_path.return_value = self.test_path
        import neon_utils.hana_utils

        def _write_token(*_, **__):
            with open(self.test_path, 'w+') as c:
                json.dump(valid_config, c)
            neon_utils.hana_utils._client_config = valid_config

        from neon_utils.hana_utils import _refresh_token
        get_token.side_effect = _write_token

        self.assertFalse(isfile(self.test_path))

        # Test valid request (auth + refresh)
        _refresh_token(self.test_server)
        get_token.assert_called_once()
        from neon_utils.hana_utils import _client_config
        self.assertTrue(isfile(self.test_path))
        with open(self.test_path) as f:
            credentials_on_disk = json.load(f)
        self.assertEqual(credentials_on_disk, _client_config)

        # Test refresh of existing token (no auth)
        sleep(1)  # sleep to ensure new credentials expire later than existing
        _refresh_token(self.test_server)
        get_token.assert_called_once()
        with open(self.test_path) as f:
            new_credentials = json.load(f)
        self.assertNotEqual(credentials_on_disk, new_credentials)
        self.assertEqual(credentials_on_disk['client_id'],
                         new_credentials['client_id'])
        self.assertEqual(credentials_on_disk['username'],
                         new_credentials['username'])
        self.assertGreater(new_credentials['expiration'],
                           credentials_on_disk['expiration'])
        self.assertNotEqual(credentials_on_disk['access_token'],
                            new_credentials['access_token'])
        self.assertNotEqual(credentials_on_disk['refresh_token'],
                            new_credentials['refresh_token'])

    def test_config_path(self):
        from neon_utils.hana_utils import _get_client_config_path
        path_1 = _get_client_config_path("https://hana.neonaialpha.com")
        default = _get_client_config_path("https://hana.neonaiservices.com")
        self.assertNotEqual(path_1, default)
        self.assertEqual(dirname(path_1), dirname(default))

        # TODO: Test invalid refresh

    @patch("ovos_config.config.Configuration")
    def test_set_default_backend_url(self, config):
        import neon_utils.hana_utils
        from neon_utils.hana_utils import set_default_backend_url
        neon_utils.hana_utils._DEFAULT_BACKEND_URL = None
        config.return_value = dict()

        set_default_backend_url()
        self.assertEqual(neon_utils.hana_utils._DEFAULT_BACKEND_URL,
                         "https://hana.neonaiservices.com")

        set_default_backend_url("https://hana.neonaialpha.com")
        self.assertEqual(neon_utils.hana_utils._DEFAULT_BACKEND_URL,
                         "https://hana.neonaialpha.com")

        set_default_backend_url()
        self.assertEqual(neon_utils.hana_utils._DEFAULT_BACKEND_URL,
                         "https://hana.neonaiservices.com")

    @patch("neon_utils.hana_utils._get_client_config_path")
    @patch("neon_utils.hana_utils._refresh_token")
    @patch("neon_utils.hana_utils.requests.post")
    def test_request_backend_ssl_verify(self, mock_post, mock_refresh, config_path):
        config_path.return_value = self.test_path

        # Mock successful response
        mock_response = unittest.mock.MagicMock()
        mock_response.ok = True
        mock_response.json.return_value = {"lang_code": "en-us", "answer": "test"}
        mock_post.return_value = mock_response

        # Use valid config to skip auth and set expiration far in future
        import neon_utils.hana_utils
        test_config = valid_config.copy() if valid_config else {
            "access_token": "test_token",
            "expiration": time() + 3600
        }
        neon_utils.hana_utils._client_config = test_config
        neon_utils.hana_utils._headers = valid_headers
        from neon_utils.hana_utils import request_backend

        # Test default SSL verification (should be True)
        request_backend("/neon/get_response",
                       {"lang_code": "en-us", "utterance": "test"},
                       self.test_server)
        mock_post.assert_called()
        call_kwargs = mock_post.call_args[1]
        self.assertTrue(call_kwargs.get('verify'))

        # Test explicit SSL verification True
        mock_post.reset_mock()
        request_backend("/neon/get_response",
                       {"lang_code": "en-us", "utterance": "test"},
                       self.test_server, ssl_verify=True)
        call_kwargs = mock_post.call_args[1]
        self.assertTrue(call_kwargs.get('verify'))

        # Test SSL verification disabled
        mock_post.reset_mock()
        request_backend("/neon/get_response",
                       {"lang_code": "en-us", "utterance": "test"},
                       self.test_server, ssl_verify=False)
        call_kwargs = mock_post.call_args[1]
        self.assertFalse(call_kwargs.get('verify'))


if __name__ == '__main__':
    unittest.main()



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Framework
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2025 Neongecko.com Inc.
# Contributors: Daniel McKnight, Guy Daniels, Elon Gasper, Richard Leeds,
# Regina Bloomstine, Casimiro Ferreira, Andrii Pernatii, Kirill Hrymailo
# BSD-3 License
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from abc import ABC, abstractmethod
from enum import IntEnum

from neon_utils.skills.neon_skill import NeonSkill
from neon_utils.logger import LOG


class CMSMatchLevel(IntEnum):
    EXACT = 1  # Skill was requested specifically
    MEDIA = 2  # Skill can handle the requested type of message (text, email, call, etc)
    LOOSE = 3  # Skill can handle a generic message request


class CommonMessageSkill(NeonSkill, ABC):
    """
    Skills that handle calls or messaging should be based on this class.
    Skills must implement 'CMS_match_query_phrase'
    This skill works with Communcation skill which handles intent matching and skill selection
    """
    def __init__(self, *args, **kwargs):
        NeonSkill.__init__(self, *args, **kwargs)

    def bind(self, bus):
        if bus:
            super().bind(bus)
            self.add_event("communication:request.message", self.__handle_send_message_request)
            self.add_event("communication:send.message", self.__handle_send_message)

            self.add_event("communication:request.call", self.__handle_place_call_request)
            self.add_event("communication:place.call", self.__handle_place_call)

    def __handle_place_call_request(self, message):
        """
        Handler for a send message request. This is the entry point for this skill to handle a request.
        :param message: message generated by Communcation skill
        """
        request_string = message.data["request"]

        # notify Communcation skill that this skill is working on it
        self.bus.emit(message.response({"request": request_string,
                                        "skill_id": self.skill_id,
                                        "searching": True}))

        # Calculate match for this skill and respond
        result: dict = self.CMS_match_call_phrase(request_string, message.context)
        if result:
            confidence = self.__calc_confidence(result, result.pop("conf"), message.context)
            LOG.debug(f"DM: Response from {self.skill_id} ({message.msg_type})")
            self.bus.emit(message.response({"request": request_string,
                                            "skill_id": self.skill_id,
                                            "conf": confidence,
                                            "skill_data": result}))
        else:
            # Notify not handling request
            self.bus.emit(message.response({"request": request_string,
                                            "skill_id": self.skill_id,
                                            "searching": False}))

    def __handle_send_message_request(self, message):
        """
        Handler for a send message request. This is the entry point for this skill to handle a request.
        :param message: message generated by Communcation skill
        """

        request_string = message.data["request"]

        # notify Communcation skill that this skill is working on it
        self.bus.emit(message.response({"request": request_string,
                                        "skill_id": self.skill_id,
                                        "searching": True}))

        # Calculate match for this skill and respond
        result: dict = self.CMS_match_message_phrase(request_string, message.context)
        if result:
            confidence = self.__calc_confidence(result, result.pop("conf"), message.context)
            LOG.debug(f"DM: Response from {self.skill_id} ({message.msg_type})")
            self.bus.emit(message.response({"request": request_string,
                                            "skill_id": self.skill_id,
                                            "conf": confidence,
                                            "skill_data": result}))
        else:
            # Notify not handling request
            self.bus.emit(message.response({"request": request_string,
                                            "skill_id": self.skill_id,
                                            "searching": False}))

    def __handle_place_call(self, message):
        """
        Callback method that handles when this skill has been selected to respond
        :param message: message object associated with request
        :return:
        """
        if message.data["skill_id"] != self.skill_id:
            # Not for this skill!
            return
        else:
            self.CMS_handle_place_call(message)

    def __handle_send_message(self, message):
        """
        Callback method that handles when this skill has been selected to respond
        :param message: message object associated with request
        :return:
        """
        if message.data["skill_id"] != self.skill_id:
            # Not for this skill!
            return
        else:
            self.CMS_handle_send_message(message)

    @staticmethod
    def __calc_confidence(result_data, level, context):
        """
        Translates a confidence level to a 0-1 value
        :param result_data (dict): data returned by skill
        :param level (CMSMatchLevel): match level
        :param context (dict): message context
        :return (float): confidence level
        """

        # Add 0.1 for each matched element (message, recipeint, subject, etc)
        bonus = (len(result_data.keys()) - 1) * 0.1

        # Lower confidence for sms requests that can't be handled
        if result_data.get("kind") == "sms" and not context.get("mobile"):
            bonus = -0.2
        # Modify confidences for phone call if mobile can handle request
        elif result_data.get("kind") == "call":
            if context.get("mobile"):
                bonus = 0.2
            else:
                bonus = -0.2

        if level == CMSMatchLevel.EXACT:
            return 1.0
        elif level == CMSMatchLevel.MEDIA:
            return 0.6 + bonus
        elif level == CMSMatchLevel.LOOSE:
            return 0.3 + bonus
        else:
            return 0.0

    @abstractmethod
    def CMS_match_message_phrase(self, request, context):
        """
        Checks if this skill can handle the given message request
        :param request: (str) user literal request to handle
        :param context: (dict) message context
        :return: (dict) confidence, data to be used in the callback if skill is selected
        """
        return {}

    @abstractmethod
    def CMS_handle_send_message(self, message):
        """
        Handles when this skill has been selected to respond
        :param message: message associated with this skills match submission
        """
        pass

    @abstractmethod
    def CMS_match_call_phrase(self, contact, context):
        """
        Checks if this skill can handle the given call request
        :param contact: (str) requested contact (name or address)
        :param context: (dict) message context
        :return: (dict) confidence, data to be used in the callback if skill is selected
        """
        return {}

    @abstractmethod
    def CMS_handle_place_call(self, message):
        """
        Handles when this skill has been selected to respond
        :param message: message associated with this skills match submission
        """
        pass

    @staticmethod
    def _extract_message_content(text):
        """
        Generic method that attempts to extract the message content from a request to send a message
        :param text: Input string to evaluate
        :return: recipient_to_append(prefix), message
        """

        if "that says " in text:
            recipient_to_append, message = text.split("that says ", 1)
        elif "saying " in text:
            recipient_to_append, message = text.split("saying ", 1)
        elif len(text) <= 1:
            message = None
            recipient_to_append = text
        else:
            recipient_to_append = None
            message = text
        return recipient_to_append, message



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Framework
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2025 Neongecko.com Inc.
# Contributors: Daniel McKnight, Guy Daniels, Elon Gasper, Richard Leeds,
# Regina Bloomstine, Casimiro Ferreira, Andrii Pernatii, Kirill Hrymailo
# BSD-3 License
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import sys
import unittest
from unittest.mock import patch

from ovos_utils.messagebus import FakeBus

import neon_utils.language_utils

sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))


class LanguageUtilTests(unittest.TestCase):
    bus = FakeBus()

    @classmethod
    def setUpClass(cls) -> None:
        import neon_utils.language_utils
        neon_utils.language_utils._bus = cls.bus

    @patch("neon_utils.language_utils.get_supported_langs")
    def test_get_supported_skills_langs(self, lf_langs):
        from neon_utils.language_utils import get_supported_skills_langs, \
            _default_lang

        lf_langs.return_value = ["en", "es", "fr", "pt", "de"]
        skill_langs = ['en', 'es', 'fr']
        native_langs = ['en']
        translate_langs = ['en', 'es', 'fr']

        def _skill_langs(msg):
            self.bus.emit(msg.response({"skill_langs": skill_langs,
                                        "native_langs": native_langs,
                                        "translate_langs": translate_langs}))

        # Test default no API response
        self.assertEqual(get_supported_skills_langs(), {_default_lang})

        # Test no translation
        self.bus.once('neon.languages.skills', _skill_langs)
        self.assertEqual(get_supported_skills_langs(False), {'en'})

        # Test with translation
        self.bus.once('neon.languages.skills', _skill_langs)
        self.assertEqual(get_supported_skills_langs(True), {'en', 'es', 'fr'})

        # Test LF unsupported lang
        translate_langs.append('na')
        skill_langs.append('na')
        self.bus.once('neon.languages.skills', _skill_langs)
        self.assertEqual(get_supported_skills_langs(True),
                         {'en', 'es', 'fr', 'na'})

        self.bus.once('neon.languages.skills', _skill_langs)
        self.assertEqual(get_supported_skills_langs(False),
                         {'en'})
        native_langs.extend(['es', 'fr'])
        self.bus.once('neon.languages.skills', _skill_langs)
        self.assertEqual(get_supported_skills_langs(False),
                         {'en', 'es', 'fr'})

    def test_get_supported_input_langs(self):
        from neon_utils.language_utils import get_supported_input_langs, \
            _default_lang

        stt_langs = ['en', 'uk-ua']

        def _stt_langs(msg):
            self.bus.emit(msg.response({"langs": stt_langs}))

        # Test default no API response
        self.assertEqual(get_supported_input_langs(skills_langs=set()),
                         {_default_lang})

        # Test skill supported languages
        self.bus.once('ovos.languages.stt', _stt_langs)
        self.assertEqual(get_supported_input_langs(skill_support=True,
                                                   skills_langs={'en', 'uk'}),
                         {'en', 'uk'})

        # Test skill unsupported languages
        self.bus.once('ovos.languages.stt', _stt_langs)
        self.assertEqual(get_supported_input_langs(skill_support=True,
                                                   skills_langs={'en', 'es'}),
                         {'en'})

        # Test skill unsupported languages
        self.bus.once('ovos.languages.stt', _stt_langs)
        self.assertEqual(get_supported_input_langs(skill_support=False,
                                                   skills_langs={'en', 'es'}),
                         {'en', 'uk'})

    def test_get_supported_output_langs(self):
        from neon_utils.language_utils import get_supported_output_langs, \
            _default_lang

        tts_langs = ['en-us', 'en-uk', 'uk-ua']

        def _tts_langs(msg):
            self.bus.emit(msg.response({"langs": tts_langs}))

        # Test default no API response
        self.assertEqual(get_supported_output_langs(skills_langs=set()),
                         {_default_lang})

        # Test skill supported languages
        self.bus.once('ovos.languages.tts', _tts_langs)
        self.assertEqual(get_supported_output_langs(skill_support=True,
                                                    skills_langs={'en', 'uk'}),
                         {'en', 'uk'})

        # Test skill unsupported languages
        self.bus.once('ovos.languages.tts', _tts_langs)
        self.assertEqual(get_supported_output_langs(skill_support=True,
                                                    skills_langs={'en', 'es'}),
                         {'en'})

        # Test skill unsupported languages
        self.bus.once('ovos.languages.tts', _tts_langs)
        self.assertEqual(get_supported_output_langs(skill_support=False,
                                                    skills_langs={'en', 'es'}),
                         {'en', 'uk'})

    def test_get_supported_languages(self):
        from neon_utils.language_utils import get_supported_languages,\
            SupportedLanguages
        tts_langs = ['en-us', 'uk-ua', 'es-mx']
        stt_langs = ['en-us', 'es-es', 'pt-pt']
        native_langs = ['en', 'na']
        translate_langs = ['en', 'es', 'fr']

        def _stt_langs(msg):
            self.bus.emit(msg.response({'langs': stt_langs}))

        def _tts_langs(msg):
            self.bus.emit(msg.response({'langs': tts_langs}))

        def _skill_langs(msg):
            self.bus.emit(msg.response({
                'native_langs': native_langs,
                'translate_langs': translate_langs,
                'skill_langs': list(set(native_langs + translate_langs))}))

        self.bus.once('ovos.languages.tts', _tts_langs)
        self.bus.once('ovos.languages.stt', _stt_langs)
        self.bus.once('neon.languages.skills', _skill_langs)
        no_translate_langs = get_supported_languages(False)
        self.assertIsInstance(no_translate_langs, SupportedLanguages)
        self.assertEqual(no_translate_langs.tts, {'en', 'uk', 'es'})
        self.assertEqual(no_translate_langs.stt, {'en', 'es', 'pt'})
        self.assertEqual(no_translate_langs.skills, {'en'})

        self.bus.once('ovos.languages.tts', _tts_langs)
        self.bus.once('ovos.languages.stt', _stt_langs)
        self.bus.once('neon.languages.skills', _skill_langs)
        with_translate_langs = get_supported_languages(True)
        self.assertIsInstance(with_translate_langs, SupportedLanguages)
        self.assertEqual(with_translate_langs.tts, {'en', 'uk', 'es'})
        self.assertEqual(with_translate_langs.stt, {'en', 'es', 'pt'})
        self.assertEqual(with_translate_langs.skills, {'en', 'es', 'fr', 'na'})


if __name__ == '__main__':
    unittest.main()



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Framework
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2025 Neongecko.com Inc.
# Contributors: Daniel McKnight, Guy Daniels, Elon Gasper, Richard Leeds,
# Regina Bloomstine, Casimiro Ferreira, Andrii Pernatii, Kirill Hrymailo
# BSD-3 License
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


from abc import ABC
from html.parser import HTMLParser


from neon_utils.logger import LOG

try:
    import requests
    from bs4 import BeautifulSoup
    from requests.exceptions import ConnectTimeout
except ImportError:
    raise ImportError("requests or bs4 not available,"
                      " pip install neon-utils[network]")


def strip_tags(html):  # TODO: Document this! DM
    class MLStripper(HTMLParser, ABC):
        def __init__(self):
            super().__init__()
            self.reset()
            self.strict = False
            self.convert_charrefs = True
            self.fed = []

        def handle_data(self, d):
            self.fed.append(d)

        def get_data(self):
            return ''.join(self.fed)

    s = MLStripper()
    s.feed(html)
    return s.get_data()


def chunks(l, n):  # TODO: Document this! DM
    return [l[i:i + n] for i in range(0, len(l), n)]


def scrape_page_for_links(url: str) -> dict:
    """
    Scrapes the passed url for any links and returns a dictionary of link labels to URLs
    :param url: Web page to scrape
    :return: Lowercase names to links on page
    """
    import unicodedata
    available_links = {}
    retry_count = 0

    def _get_links(url):
        LOG.debug(url)
        try:
            html = requests.get(url, timeout=2.0).text
        except ConnectTimeout as e:
            raise e
        except Exception as e:
            LOG.warning(e)
            html = None
        if not str(url).startswith("http") and not html:
            request_url = f"https://{url}"
            try:
                html = requests.get(request_url, timeout=2.0).text
            except ConnectTimeout as e:
                raise e
            except Exception as e:
                LOG.warning(e)
                html = None
            if not html:
                try:
                    request_url = f"http://{url}"
                    html = requests.get(request_url, timeout=2.0).text
                except ConnectTimeout as e:
                    raise e
                except Exception as e:
                    LOG.warning(e)
                    html = None
            url = request_url

        LOG.debug(url)
        soup = BeautifulSoup(html, 'lxml')
        # LOG.debug(html)
        # LOG.debug(soup)

        # Look through the page and find all anchor tags
        for i in soup.find_all("a", href=True):
            # LOG.debug(f"DM: found link: {i.text.rstrip()}")
            # LOG.debug(f"DM: found href: {i['href']}")

            if '://' not in i['href']:
                # Assume this is a relative address
                href = url + i['href'].lower()
            elif url.split('://')[1] in i['href']:
                href = i['href'].lower()
            else:
                href = None

            if href:
                link_name = (
                    unicodedata.normalize("NFKD",
                                          i.text.rstrip().replace(u"\u2013", "")
                                          .replace(u"\u201d", "")
                                          .replace(u"\u201c", "")
                                          .replace('"', "")
                                          .replace("'", "")
                                          .replace("&apos;", "")
                                          .replace("\n", "").lower()))
                available_links[link_name] = href
                LOG.debug(f"found link={link_name}|href={href}")

        LOG.debug(available_links)

    try:
        _get_links(url)
    except ConnectTimeout:
        retry_count += 1
        if retry_count < 8:
            _get_links(url)
        else:
            raise ConnectTimeout
    except Exception as x:
        LOG.error(x)
        LOG.debug(available_links)
        raise ReferenceError
    return available_links
